@@ -1,6 +1,19 @@
 #ifndef IIOP_PRIVATE_H
 #define IIOP_PRIVATE_H 1
 
+
+#include "config.h"
+
+#ifdef HAVE_SYS_POLL_H
+#include <sys/poll.h>
+#endif
+
+#include <sys/time.h>
+#include <sys/types.h>
+#include <unistd.h>
+
+#include <glib.h>
+
 typedef struct {
   GList *list;
   gboolean connection_list_changed;
