@@ -1,7 +1,7 @@
 #ifndef THREAD_SAFETY_H
 #define THREAD_SAFETY_H 1
 
-#ifdef _REENTRANT
+#ifdef NOT_REENTRANT
 
 #include <pthread.h>
 
@@ -18,6 +18,7 @@ extern pthread_key_t thread_data;
 
 #else
 
+/* stupid work around ANSI & empty semicolons. */
 #define DEFINE_LOCK(x)
 #define INIT_LOCK(x)
 #define GET_LOCK(x)
