@@ -29,28 +29,120 @@
 #include <assert.h>
 
 #include "orbit.h"
+#include "orbit_poa.h"
+#include "genrand.h"
 
-/* Stupid inheritance... :) we have to do the same thing five times or so,
-   might as well make a macro for it. */
+PortableServer_ThreadPolicyValue
+PortableServer_ThreadPolicy__get_value(PortableServer_ThreadPolicy  obj, CORBA_Environment *ev)
+{
+	if(!obj) {
+		ev->_major = 2 ;
+		goto error_exit;
+	}
+	ev->_major = CORBA_NO_EXCEPTION;
 
-#define DO_POLICY_GET_VALUE(x) PortableServer_##x##Value PortableServer_##x##__get_value(PortableServer_##x obj, CORBA_Environment *ev) { \
-if(!obj) { \
-ev->_major = ex_CORBA_BAD_PARAM; \
-goto error_exit; \
-} \
-return obj->value; \
-error_exit: \
-CORBA_exception_set_system(ev, 0, CORBA_COMPLETED_NO); \
-return 0; \
+	return obj->value;
+
+error_exit:
+	CORBA_exception_set_system(ev, 0, CORBA_COMPLETED_NO);
+	return 0;
 }
 
-DO_POLICY_GET_VALUE(ThreadPolicy)
-DO_POLICY_GET_VALUE(LifespanPolicy)
-DO_POLICY_GET_VALUE(IdUniquenessPolicy)
-DO_POLICY_GET_VALUE(IdAssignmentPolicy)
-DO_POLICY_GET_VALUE(ImplicitActivationPolicy)
-DO_POLICY_GET_VALUE(ServantRetentionPolicy)
-DO_POLICY_GET_VALUE(RequestProcessingPolicy)
+PortableServer_LifespanPolicyValue
+PortableServer_LifespanPolicy__get_value(PortableServer_LifespanPolicy  obj, CORBA_Environment *ev)
+{
+	if(!obj) {
+		ev->_major = 2 ;
+		goto error_exit;
+	}
+
+	ev->_major = CORBA_NO_EXCEPTION;
+	return obj->value;
+
+error_exit:
+	CORBA_exception_set_system(ev, 0, CORBA_COMPLETED_NO);
+	return 0;
+}
+
+PortableServer_IdUniquenessPolicyValue
+PortableServer_IdUniquenessPolicy__get_value(PortableServer_IdUniquenessPolicy  obj, CORBA_Environment *ev)
+{
+	if(!obj) {
+		ev->_major = 2 ;
+		goto error_exit;
+	}
+
+	ev->_major = CORBA_NO_EXCEPTION;
+	return obj->value;
+
+error_exit:
+	CORBA_exception_set_system(ev, 0, CORBA_COMPLETED_NO);
+	return 0;
+}
+
+PortableServer_IdAssignmentPolicyValue
+PortableServer_IdAssignmentPolicy__get_value(PortableServer_IdAssignmentPolicy  obj, CORBA_Environment *ev)
+{
+	if(!obj) {
+		ev->_major = 2 ;
+		goto error_exit;
+	}
+
+	ev->_major = CORBA_NO_EXCEPTION;
+	return obj->value;
+
+error_exit:
+	CORBA_exception_set_system(ev, 0, CORBA_COMPLETED_NO);
+	return 0;
+}
+
+PortableServer_ImplicitActivationPolicyValue
+PortableServer_ImplicitActivationPolicy__get_value(PortableServer_ImplicitActivationPolicy  obj, CORBA_Environment *ev)
+{
+	if(!obj) {
+		ev->_major = 2 ;
+		goto error_exit;
+	}
+
+	ev->_major = CORBA_NO_EXCEPTION;
+	return obj->value;
+
+error_exit:
+	CORBA_exception_set_system(ev, 0, CORBA_COMPLETED_NO);
+	return 0;
+}
+
+PortableServer_ServantRetentionPolicyValue
+PortableServer_ServantRetentionPolicy__get_value(PortableServer_ServantRetentionPolicy  obj, CORBA_Environment *ev)
+{
+	if(!obj) {
+		ev->_major = 2 ;
+		goto error_exit;
+	}
+
+	ev->_major = CORBA_NO_EXCEPTION;
+	return obj->value;
+
+error_exit:
+	CORBA_exception_set_system(ev, 0, CORBA_COMPLETED_NO);
+	return 0;
+}
+
+PortableServer_RequestProcessingPolicyValue
+PortableServer_RequestProcessingPolicy__get_value(PortableServer_RequestProcessingPolicy  obj, CORBA_Environment *ev)
+{
+	if(!obj) {
+		ev->_major = 2 ;
+		goto error_exit;
+	}
+
+	ev->_major = CORBA_NO_EXCEPTION;
+	return obj->value;
+
+error_exit:
+	CORBA_exception_set_system(ev, 0, CORBA_COMPLETED_NO);
+	return 0;
+}
 
 /* make emacs happy; */
 
@@ -65,6 +157,7 @@ PortableServer_POAManager_get_state(PortableServer_POAManager obj,
 		return -1;
 	}
 
+	ev->_major = CORBA_NO_EXCEPTION;
 	return obj->state;
 }
 
@@ -113,9 +206,12 @@ PortableServer_POAManager_activate(PortableServer_POAManager obj,
 
 		g_slist_foreach(todo, (GFunc)ORBit_POA_handle_request,
 				curpoa);
+		g_slist_foreach(todo, (GFunc)giop_recv_buffer_unuse,
+				NULL);
 
 		g_slist_free(todo);
 	}
+	ev->_major = CORBA_NO_EXCEPTION;
 }
 
 void
@@ -140,6 +236,8 @@ PortableServer_POAManager_hold_requests(PortableServer_POAManager obj,
 	obj->state = PortableServer_POAManager_HOLDING;
 	if(!wait_for_completion)
 		g_warning("hold_requests not finished - don't know how to kill outstanding request fulfillments");
+
+	ev->_major = CORBA_NO_EXCEPTION;
 }
 
 void
@@ -164,6 +262,7 @@ PortableServer_POAManager_discard_requests(PortableServer_POAManager obj,
 	obj->state = PortableServer_POAManager_DISCARDING;
 	if(!wait_for_completion)
 		g_warning("discard_requests not finished - don't know how to kill outstanding request fulfillments");
+	ev->_major = CORBA_NO_EXCEPTION;
 }
 
 void
@@ -190,6 +289,7 @@ PortableServer_POAManager_deactivate(PortableServer_POAManager obj,
 
 	if(etherealize_objects)
 		g_slist_foreach(obj->poa_collection, (GFunc)ORBit_POA_etherealize_objects, ev);
+	ev->_major = CORBA_NO_EXCEPTION;
 }
 
 
@@ -240,16 +340,27 @@ PortableServer_POA_create_POA
 		CORBA_Environment *ev)
 {
 	PortableServer_POA new_poa = NULL;
+	PortableServer_POA check_poa = NULL;
+	
+	/* Check for a child POA by the same name in parent */
+	check_poa = PortableServer_POA_find_POA(poa,adapter_name,
+						FALSE, ev);
+	CORBA_exception_free (ev);
 
-	if(policies)
-		g_warning("non-default POA policies aren't implemented yet");
+	if (!check_poa) {
+	  new_poa = ORBit_POA_new(poa->orb,
+				  adapter_name, a_POAManager, policies, ev);
+	} else {
+	  CORBA_exception_set(ev, CORBA_USER_EXCEPTION,
+			  ex_PortableServer_POA_AdapterAlreadyExists,
+			  NULL);
+	  new_poa = NULL;
+	}
 
-	new_poa = ORBit_POA_new(poa->orb,
-				adapter_name, a_POAManager, policies, ev);
-	new_poa->parent_poa = poa;
-
-	if(ev->_major == CORBA_NO_EXCEPTION)
-		ORBit_POA_add_child(poa, new_poa, ev);
+	if(ev->_major == CORBA_NO_EXCEPTION) {
+	  new_poa->parent_poa = poa;
+	  ORBit_POA_add_child(poa, new_poa, ev);
+	}
 
 	return new_poa;
 }
@@ -278,8 +389,10 @@ PortableServer_POA_find_POA(PortableServer_POA obj,
 	for(curitem = obj->child_POAs; curitem;
 	    curitem = g_slist_next(curitem)) {
 		child_poa = (PortableServer_POA)curitem->data;
-		if(!strcmp(child_poa->the_name, adapter_name))
+		if(!strcmp(child_poa->the_name, adapter_name)) {
+			ev->_major = CORBA_NO_EXCEPTION;
 			return child_poa;
+		}
 	}
 
 	if(activate_it)
@@ -315,6 +428,7 @@ PortableServer_POA_destroy(PortableServer_POA obj,
 			  ORBIT_ROOT_OBJECT(obj)->refs);
 
 	CORBA_Object_release((CORBA_Object)obj, ev);
+	ev->_major = CORBA_NO_EXCEPTION;
 }
 
 PortableServer_ThreadPolicy PortableServer_POA_create_thread_policy(PortableServer_POA obj, PortableServer_ThreadPolicyValue value, CORBA_Environment *ev)
@@ -540,7 +654,26 @@ void PortableServer_POA_set_servant(PortableServer_POA obj, PortableServer_Serva
 	obj->default_servant = p_servant;
 }
 
-CORBA_ORB
+static CORBA_unsigned_long
+get_objnum_for_obj(PortableServer_POA poa, ORBit_POAObject *obj)
+{
+	CORBA_unsigned_long retval;
+
+	if(poa->first_free_id) {
+		retval = poa->first_free_id;
+		poa->first_free_id = GPOINTER_TO_UINT(g_ptr_array_index(poa->objnum_to_obj,
+									retval));
+		g_ptr_array_index(poa->objnum_to_obj, retval) = obj;
+	} else {
+		retval = poa->objnum_to_obj->len;
+		g_ptr_array_add(poa->objnum_to_obj,
+				obj);
+	}
+
+	return retval;
+}
+
+static CORBA_ORB
 get_orb_for_poa(PortableServer_POA poa)
 {
 	if(poa->orb)
@@ -569,28 +702,43 @@ PortableServer_POA_activate_object(PortableServer_POA obj,
 				    NULL);
 		return NULL;
 	}
+	
+	/* Servant Already Active */
+	if((obj->id_uniqueness==PortableServer_UNIQUE_ID) &&
+	   (ORBIT_OBJECT_KEY(servant->_private)->object != 0)) {
+		CORBA_exception_set(ev, CORBA_USER_EXCEPTION,
+				    ex_PortableServer_POA_ServantAlreadyActive,
+				    NULL);
+		return NULL;
+	}
+	
 
-	new_obj = g_new(ORBit_POAObject, 1);
+	new_obj = g_new0(ORBit_POAObject, 1);
 	new_obj->object_id = (PortableServer_ObjectId*)CORBA_sequence_octet__alloc();
 
 	new_objid =
 		ORBit_POA_allocate_oid(obj,
-				      ORBIT_OBJECT_KEY(servant->_private)->class_info->class_name);
-
+					ORBIT_OBJECT_KEY(servant->_private)->class_info->class_name);
+	
 	new_obj->object_id->_buffer = CORBA_octet_allocbuf(new_objid->_length);
 	new_obj->object_id->_length = new_objid->_length;
 	memcpy(new_obj->object_id->_buffer, new_objid->_buffer,
-	       new_objid->_length);
+		new_objid->_length);
 	CORBA_sequence_set_release(new_obj->object_id, CORBA_TRUE);
 
 	new_obj->servant = p_servant;
 	ORBIT_OBJECT_KEY(servant->_private)->object = new_obj;
 	new_obj->orb = get_orb_for_poa(obj);
+	new_obj->poa = obj;
+	new_obj->objnum = get_objnum_for_obj(obj, new_obj);
+	orbit_genrand(new_obj->rand_data, ORBIT_RAND_KEY_LEN);
 
 	g_hash_table_insert(obj->active_object_map,
 			    new_obj->object_id,
 			    new_obj);
 
+	ev->_major = CORBA_NO_EXCEPTION;
+
 	return new_objid;
 }
 
@@ -603,9 +751,7 @@ PortableServer_POA_activate_object_with_id(PortableServer_POA obj,
 	PortableServer_ServantBase *servant = p_servant;
 	ORBit_POAObject *newobj;
 
-	if(!obj || !id || !p_servant
-	   || id->_buffer[id->_length - 1]
-	   || strchr(id->_buffer, '/')) {
+	if(!obj || !id || !p_servant) {
 		CORBA_exception_set_system(ev, ex_CORBA_BAD_PARAM,
 					   CORBA_COMPLETED_NO);
 		return;
@@ -614,17 +760,23 @@ PortableServer_POA_activate_object_with_id(PortableServer_POA obj,
 	newobj = g_hash_table_lookup(obj->active_object_map,
 				     id);
 
-	if(!newobj) {
-		newobj = g_new(ORBit_POAObject, 1);
-		newobj->object_id = (PortableServer_ObjectId *)CORBA_sequence_octet__alloc();
-		newobj->object_id->_length = id->_length;
-		newobj->object_id->_buffer = CORBA_octet_allocbuf(id->_length);
-		newobj->object_id->_release = CORBA_TRUE;
-		memcpy(newobj->object_id->_buffer, id->_buffer, id->_length);
-		newobj->poa = obj;
-		newobj->orb = get_orb_for_poa(obj);
+	if(newobj) {
+		CORBA_exception_set(ev, CORBA_USER_EXCEPTION,
+				    ex_PortableServer_POA_ObjectAlreadyActive, NULL);
+		return;
 	}
 
+	newobj = g_new0(ORBit_POAObject, 1);
+	newobj->object_id = (PortableServer_ObjectId *)CORBA_sequence_octet__alloc();
+	newobj->object_id->_length = id->_length;
+	newobj->object_id->_buffer = CORBA_octet_allocbuf(id->_length);
+	newobj->object_id->_release = CORBA_TRUE;
+	memcpy(newobj->object_id->_buffer, id->_buffer, id->_length);
+	newobj->poa = obj;
+	newobj->orb = get_orb_for_poa(obj);
+	newobj->objnum = get_objnum_for_obj(obj, newobj);
+	orbit_genrand(newobj->rand_data, ORBIT_RAND_KEY_LEN);
+
 	newobj->servant = p_servant;
 
 	g_hash_table_insert(obj->active_object_map,
@@ -632,6 +784,8 @@ PortableServer_POA_activate_object_with_id(PortableServer_POA obj,
 			    newobj);
 
 	ORBIT_OBJECT_KEY(servant->_private)->object = newobj;
+
+	ev->_major = CORBA_NO_EXCEPTION;
 }
 
 void
@@ -657,6 +811,9 @@ PortableServer_POA_deactivate_object(PortableServer_POA obj,
 		return;
 	}
 
+	g_ptr_array_index(obj->objnum_to_obj, oldobj->objnum) = GUINT_TO_POINTER(obj->first_free_id);
+	obj->first_free_id = oldobj->objnum;
+
 	g_hash_table_remove(obj->active_object_map, oid);
 
 	if(obj->request_processing == PortableServer_USE_SERVANT_MANAGER) {
@@ -666,15 +823,19 @@ PortableServer_POA_deactivate_object(PortableServer_POA obj,
 		sm = (POA_PortableServer_ServantActivator *)obj->servant_manager;
 		epv = sm->vepv->PortableServer_ServantActivator_epv;
 		epv->etherealize(sm, oldobj->object_id, obj,
-				 oldobj->servant,
-				 CORBA_FALSE,
-				 CORBA_FALSE,
-				 ev);
+				oldobj->servant,
+				CORBA_FALSE,
+				CORBA_FALSE,
+				ev);
+	} else {
+	  PortableServer_ServantBase *serv = oldobj->servant;
+	  serv->_private = NULL;
 	}
 
 	CORBA_free(oldobj->object_id);
 
 	g_free(oldobj);
+	ev->_major = CORBA_NO_EXCEPTION;
 }
 
 CORBA_Object
@@ -686,58 +847,117 @@ PortableServer_POA_create_reference(PortableServer_POA obj,
 	return(NULL);
 }
 
-CORBA_Object PortableServer_POA_create_reference_with_id(PortableServer_POA obj, PortableServer_ObjectId *oid, CORBA_RepositoryId intf, CORBA_Environment *ev)
-{
-	g_assert(!"Not yet implemented");
-	return(NULL);
-}
-
 PortableServer_ObjectId *PortableServer_POA_servant_to_id(PortableServer_POA obj, PortableServer_Servant p_servant, CORBA_Environment *ev)
 {
+	PortableServer_ObjectId *retval, *orig;
 	PortableServer_ServantBase *serv = p_servant;
 	g_return_val_if_fail(p_servant != NULL, NULL);
-	return ORBIT_OBJECT_KEY(serv->_private)->object->object_id;
-}
 
-CORBA_Object PortableServer_POA_servant_to_reference(PortableServer_POA obj, PortableServer_Servant p_servant, CORBA_Environment *ev)
-{
-	CORBA_Object retval;
-
-	retval = PortableServer_POA_id_to_reference(obj,
-						    ORBIT_OBJECT_KEY(((PortableServer_ServantBase*)p_servant)->_private)->object->object_id, ev);
+	orig = ORBIT_OBJECT_KEY(serv->_private)->object->object_id;
+	retval = (PortableServer_ObjectId *)CORBA_sequence_octet__alloc();
+	retval->_length = retval->_maximum = orig->_length;
+	retval->_buffer = CORBA_octet_allocbuf(retval->_length);
+	memcpy(retval->_buffer, orig->_buffer, retval->_length);
+	CORBA_sequence_set_release(retval, CORBA_TRUE);
 
 	return retval;
 }
 
-PortableServer_Servant PortableServer_POA_reference_to_servant(PortableServer_POA obj, CORBA_Object reference, CORBA_Environment *ev)
+CORBA_Object
+PortableServer_POA_servant_to_reference(PortableServer_POA obj, PortableServer_Servant p_servant, CORBA_Environment *ev)
 {
-	ORBit_POAObject *objinfo;
+	CORBA_Object retval;
+	PortableServer_ObjectId *orig_id;
+	PortableServer_ServantBase *servant = p_servant;
+	ORBit_ObjectKey *obj_key =  ORBIT_OBJECT_KEY(servant->_private);
+	
+	int implicit = (obj->implicit_activation == PortableServer_IMPLICIT_ACTIVATION);
+	int activate_able = (obj_key->object == 0) ||
+		(obj->id_uniqueness==PortableServer_MULTIPLE_ID);
+	/* ImplicitActivationPolicy */
+	if( implicit && activate_able) {
+		orig_id = PortableServer_POA_activate_object(obj, p_servant, ev);
+	} else {
+		orig_id = obj_key->object->object_id;
+	}
+	retval = PortableServer_POA_id_to_reference(obj,orig_id,ev);
 
-	g_assert(!"NYI");
+	if( implicit && activate_able)
+		CORBA_free (orig_id);
+
+	return retval;
+}
+
+PortableServer_Servant
+PortableServer_POA_reference_to_servant(PortableServer_POA obj, CORBA_Object reference, CORBA_Environment *ev)
+{
+	GSList *cur;
+
+	g_assert(reference);
 
 	if(obj->request_processing != PortableServer_USE_DEFAULT_SERVANT
 	   && obj->servant_retention != PortableServer_RETAIN) {
 		CORBA_exception_set(ev, CORBA_USER_EXCEPTION,
 				    ex_PortableServer_POA_WrongPolicy,
 				    NULL);
+		return NULL;
 	}
 
-	objinfo = g_hash_table_lookup(obj->active_object_map,
-				      strrchr(reference->objinfo.iopinfo->object_key._buffer, '/') + 1);
+	if(reference->servant)
+		return reference->servant;
 
-	if(!objinfo) {
-		CORBA_exception_set(ev, CORBA_USER_EXCEPTION,
-				    ex_PortableServer_POA_ObjectNotActive,
-				    NULL);
+	for(cur = reference->profile_list; cur; cur = cur->next) {
+		PortableServer_ObjectId *oid;
+		ORBit_Object_info *curprof = cur->data;
+		ORBit_POAObject *objinfo;
+
+		objinfo = ORBit_POA_find_oid_for_object_key(obj, &(curprof->object_key), &oid);
+		CORBA_free(oid);
+		if(objinfo)
+			return objinfo->servant;
 	}
 
-	return objinfo->servant;
+	CORBA_exception_set(ev, CORBA_USER_EXCEPTION,
+			    ex_PortableServer_POA_ObjectNotActive,
+			    NULL);
+	return NULL;
 }
 
 PortableServer_ObjectId *PortableServer_POA_reference_to_id(PortableServer_POA obj, CORBA_Object reference, CORBA_Environment *ev)
 {
-	g_assert(!"Not yet implemented");
-	return(NULL);
+	PortableServer_ObjectId *retval;
+	ORBit_POAObject *objinfo;
+
+	g_assert(reference);
+
+	ORBit_object_get_connection(reference);
+
+	g_assert(reference->active_profile);
+
+	if(obj->request_processing != PortableServer_USE_DEFAULT_SERVANT
+	   && obj->servant_retention != PortableServer_RETAIN) {
+		CORBA_exception_set(ev, CORBA_USER_EXCEPTION,
+				    ex_PortableServer_POA_WrongPolicy,
+				    NULL);
+		return NULL;
+	}
+
+	objinfo = ORBit_POA_find_oid_for_object_key(obj, &(reference->active_profile->object_key), &retval);
+	if(objinfo) {
+		CORBA_free(retval);
+		retval = (PortableServer_ObjectId *)CORBA_sequence_octet__alloc();
+		retval->_length = retval->_maximum = objinfo->object_id->_length;
+		retval->_buffer = CORBA_octet_allocbuf(retval->_length);
+		memcpy(retval->_buffer, objinfo->object_id->_buffer, retval->_length);
+		CORBA_sequence_set_release(retval, CORBA_TRUE);
+		return retval;
+	} else if(retval)
+		return retval;
+
+	CORBA_exception_set(ev, CORBA_USER_EXCEPTION,
+			    ex_PortableServer_POA_ObjectNotActive,
+			    NULL);
+	return NULL;
 }
 
 PortableServer_Servant PortableServer_POA_id_to_servant(PortableServer_POA obj, PortableServer_ObjectId *oid, CORBA_Environment *ev)
@@ -764,72 +984,105 @@ PortableServer_Servant PortableServer_POA_id_to_servant(PortableServer_POA obj,
 	return objinfo->servant;
 }
 
-CORBA_Object PortableServer_POA_id_to_reference(PortableServer_POA obj,
-						PortableServer_ObjectId *oid,
-						CORBA_Environment *ev)
+static CORBA_Object
+my_PortableServer_POA_id_to_reference(PortableServer_POA obj,
+				      PortableServer_ObjectId *oid,
+				      const char *type_id,
+				      CORBA_Environment *ev)
 {
-	GHashTable *objinfo;
+	GSList *profiles=NULL;
+	ORBit_Object_info *object_info;
 	CORBA_Object retval;
 	CORBA_ORB orb;
 	ORBit_POAObject *pobj;
-	TAG_INTERNET_IOP_info iopinfo = {1,0,NULL,0,{0,0,NULL,0}};
-	TAG_ORBIT_SPECIFIC_info orbitinfo = { "", 0 };
 	ORBit_ObjectKey *objkey = NULL;
 
-	objinfo = g_hash_table_new(g_str_hash, g_str_equal);
-
 	orb = obj->the_POAManager->orb;
 
 	g_assert(!oid->_buffer[oid->_length - 1]);
 
 	pobj = g_hash_table_lookup(obj->active_object_map, oid);
 
-	if(pobj && pobj->servant) {
+	if(pobj) {
 		objkey = ORBIT_OBJECT_KEY(((PortableServer_ServantBase *)pobj->servant)->_private);
-		g_hash_table_insert(objinfo, "type_id",
-				    (gpointer)objkey->class_info->class_name);
-
-	} else {
-		g_hash_table_insert(objinfo, "type_id", "Null");
+		type_id= objkey->class_info->class_name;
 	}
-	ORBit_POA_find_object_key_for_oid(obj, oid, &iopinfo.object_key);
-	if(orb->cnx.ipv4) {
-		iopinfo.host = IIOP_CONNECTION(orb->cnx.ipv4)->u.ipv4.hostname;
-		iopinfo.port = ntohs(IIOP_CONNECTION(orb->cnx.ipv4)->u.ipv4.location.sin_port);
-	}
-	g_hash_table_insert(objinfo, "IOP_TAG_INTERNET_IOP", &iopinfo);
 
+	/* Do the local connection first, so it will be attempted first by
+	   the client parsing the IOR string
+	 */
 	if(orb->cnx.ipv6 || orb->cnx.usock) {
+		object_info = g_new0(ORBit_Object_info, 1);
+
+		object_info->profile_type=IOP_TAG_ORBIT_SPECIFIC;
+		object_info->iiop_major = 1;
+		object_info->iiop_minor = 0;
+
+		ORBit_POA_find_object_key_for_oid(obj, pobj, oid, &object_info->object_key);
+
 #ifdef HAVE_IPV6
 		if(orb->cnx.ipv6) {
-			orbitinfo.ipv6_port =
+			object_info->tag.orbitinfo.ipv6_port =
 				ntohs(IIOP_CONNECTION(orb->cnx.ipv6)->u.ipv6.location.sin_port);
 		}
 #endif
 		if(orb->cnx.usock) {
-			orbitinfo.unix_sock_path =
-				IIOP_CONNECTION(orb->cnx.usock)->u.usock.sun_path;
+			object_info->tag.orbitinfo.unix_sock_path =
+				g_strdup(IIOP_CONNECTION(orb->cnx.usock)->u.usock.sun_path);
 		}
-		g_hash_table_insert(objinfo, "IOP_TAG_ORBIT_SPECIFIC", &orbitinfo);
+		ORBit_set_object_key(object_info);
+		profiles=g_slist_append(profiles, object_info);
 	}
 
-	retval = ORBit_create_object_with_info(objinfo, orb, ev);
+	if(orb->cnx.ipv4) {
+		object_info=g_new0(ORBit_Object_info, 1);
 
-	g_hash_table_destroy(objinfo);
-	CORBA_free(iopinfo.object_key._buffer);
+		object_info->profile_type = IOP_TAG_INTERNET_IOP;
+		object_info->iiop_major = 1;
+		object_info->iiop_minor = 0;
+		ORBit_POA_find_object_key_for_oid(obj, pobj, oid, &object_info->object_key);
+
+		object_info->tag.iopinfo.host = g_strdup(IIOP_CONNECTION(orb->cnx.ipv4)->u.ipv4.hostname);
+		object_info->tag.iopinfo.port = ntohs(IIOP_CONNECTION(orb->cnx.ipv4)->u.ipv4.location.sin_port);
+
+		ORBit_set_object_key(object_info);
+		profiles=g_slist_append(profiles, object_info);
+	}
+
+	retval = ORBit_create_object_with_info(profiles, type_id, orb, ev);
 
 	if(retval != CORBA_OBJECT_NIL
 	   && ev->_major == CORBA_NO_EXCEPTION
-	   && objkey) {
+	   && objkey && objkey->class_info && objkey->class_info->init_local_objref) {
+	  if(!retval->vepv)
+	    {
+	     /* XXX potential memleak if we get an already-valid objref */
 		retval->vepv = g_new0(gpointer, ORBit_class_assignment_counter + 1);
+		retval->vepv_size = ORBit_class_assignment_counter + 1;
 		objkey->class_info->init_local_objref(retval, pobj->servant);
 		retval->servant = pobj->servant;
+	    }
 	} else
 		retval->vepv = retval->servant = NULL;
 
 	return retval;
 }
 
+CORBA_Object PortableServer_POA_id_to_reference(PortableServer_POA obj,
+						PortableServer_ObjectId *oid,
+						CORBA_Environment *ev)
+{
+	return my_PortableServer_POA_id_to_reference(obj, oid, NULL, ev);
+}
+
+CORBA_Object
+PortableServer_POA_create_reference_with_id(PortableServer_POA obj,
+					    PortableServer_ObjectId *oid,
+					    CORBA_RepositoryId intf,
+					    CORBA_Environment *ev)
+{
+	return my_PortableServer_POA_id_to_reference(obj, oid, intf, ev);
+}
 
 PortableServer_POA PortableServer_Current_get_POA(PortableServer_Current obj, CORBA_Environment *ev)
 {
@@ -846,7 +1099,7 @@ PortableServer_ObjectId *PortableServer_Current_get_object_id(PortableServer_Cur
 
 CORBA_char *PortableServer_ObjectId_to_string(PortableServer_ObjectId *id, CORBA_Environment *env)
 {
-	return CORBA_string_dup(id->_buffer);
+	return CORBA_string_dup((CORBA_char *)id->_buffer);
 }
 
 CORBA_wchar *PortableServer_ObjectId_to_wstring(PortableServer_ObjectId *id, CORBA_Environment *env)
@@ -919,72 +1172,228 @@ void PortableServer_ServantBase__fini(PortableServer_Servant servant,
 
 CORBA_Identifier CORBA_ServerRequest_operation(CORBA_ServerRequest req, CORBA_Environment *env)
 {
-	g_assert(!"Not yet implemented");
-	return(NULL);
+	return CORBA_string_dup(req->rbuf->message.u.request.operation);
 }
 
-CORBA_Context *CORBA_ServerRequest_ctx(CORBA_ServerRequest req, CORBA_Environment *env)
+CORBA_Context
+CORBA_ServerRequest_ctx(CORBA_ServerRequest req, CORBA_Environment *env)
 {
-	g_assert(!"Not yet implemented");
-	return(NULL);
+	if(!req->params || req->did_ctx) {
+		CORBA_exception_set_system(env, ex_CORBA_BAD_INV_ORDER,
+					   CORBA_COMPLETED_NO);
+		return NULL;
+	}
+
+	return NULL;
 }
 
-void CORBA_ServerRequest_arguments(CORBA_ServerRequest req, CORBA_NVList *parameters, CORBA_Environment *env)
+void
+CORBA_ServerRequest_arguments(CORBA_ServerRequest req,
+			      CORBA_NVList *parameters,
+			      CORBA_Environment *env)
 {
-	g_assert(!"Not yet implemented");
-	return;
+	int i;
+
+	if(req->params) {
+		CORBA_exception_set_system(env, ex_CORBA_BAD_INV_ORDER,
+					   CORBA_COMPLETED_NO);
+		return;
+	}
+
+	req->params = parameters;
+
+	for(i = 0; i < parameters->list->len; i++) {
+		CORBA_NamedValue *cur;
+
+		cur = &g_array_index(parameters->list, CORBA_NamedValue, i);
+
+		if(cur->arg_modes & CORBA_ARG_OUT) continue;
+		cur->argument._value = ORBit_demarshal_arg(req->rbuf,
+							   cur->argument._type,
+							   TRUE,
+							   (CORBA_ORB)req->orb);
+		CORBA_any_set_release(&cur->argument, TRUE);
+	}
 }
 
-void CORBA_ServerRequest_set_result(CORBA_ServerRequest req, CORBA_any *value, CORBA_Environment *env)
+void
+CORBA_ServerRequest_set_result(CORBA_ServerRequest req,
+			       CORBA_any *value,
+			       CORBA_Environment *env)
 {
-	g_assert(!"Not yet implemented");
-	return;
+	if(req->sbuf) {
+		CORBA_exception_set_system(env, ex_CORBA_BAD_INV_ORDER,
+					   CORBA_COMPLETED_NO);
+		return;
+	}
+
+	req->sbuf = giop_send_reply_buffer_use(GIOP_MESSAGE_BUFFER(req->rbuf)->connection,
+					       NULL,
+					       req->rbuf->message.u.request.request_id,
+					       CORBA_NO_EXCEPTION);
+	if(!req->sbuf) {
+		CORBA_exception_set_system(env, ex_CORBA_COMM_FAILURE,
+					   CORBA_COMPLETED_NO);
+		return;
+	}
+		
+	ORBit_marshal_arg(req->sbuf, value->_value, value->_type);
 }
 
-void CORBA_ServerRequest_set_exception(CORBA_ServerRequest req, CORBA_exception_type major, CORBA_any *value, CORBA_Environment *env)
+void
+CORBA_ServerRequest_set_exception(CORBA_ServerRequest req,
+				  CORBA_exception_type major,
+				  CORBA_any *value,
+				  CORBA_Environment *env)
 {
-	g_assert(!"Not yet implemented");
-	return;
+	if(req->sbuf) {
+		CORBA_exception_set_system(env, ex_CORBA_BAD_INV_ORDER,
+					   CORBA_COMPLETED_NO);
+		return;
+	}
+
+	req->sbuf = giop_send_reply_buffer_use(GIOP_MESSAGE_BUFFER(req->rbuf)->connection,
+					       NULL,
+					       req->rbuf->message.u.request.request_id,
+					       major);
+	if(!req->sbuf) {
+		CORBA_exception_set_system(env, ex_CORBA_COMM_FAILURE,
+					   CORBA_COMPLETED_NO);
+		return;
+	}
+
+	req->did_exc = TRUE;
+
+	/* XXX do we really need to copy the repo_id into the
+	   send buffer? Or is there a way to assume that the CORBA_TypeCode
+	   value->_type will be around until after we send the message? */
+	{
+		CORBA_unsigned_long slen;
+		slen = strlen(value->_type->repo_id) + 1;
+		giop_send_buffer_append_mem_indirect_a(req->sbuf, &slen,
+						       sizeof(slen));
+		giop_send_buffer_append_mem_indirect(req->sbuf,
+						     value->_type->repo_id,
+						     slen);
+	}
+		
+	ORBit_marshal_arg(req->sbuf, value->_value, value->_type);
 }
 
 void
 POA_PortableServer_ServantActivator__init(PortableServer_Servant servant,
-					  CORBA_Environment * ev)
+					CORBA_Environment * ev)
 {
-   static const PortableServer_ClassInfo class_info =
-   {NULL,
-    "IDL:omg.org/PortableServer/ServantActivator:1.0",
-    NULL };
+	static const PortableServer_ClassInfo class_info =
+	{NULL,
+	"IDL:omg.org/PortableServer/ServantActivator:1.0",
+	NULL};
 
-   PortableServer_ServantBase__init(((PortableServer_ServantBase *) servant), ev);
+	PortableServer_ServantBase__init(((PortableServer_ServantBase *) servant), ev);
 
-   ORBIT_OBJECT_KEY(((PortableServer_ServantBase *)servant)->_private)->class_info = (gpointer)&class_info;
+	ORBIT_OBJECT_KEY(((PortableServer_ServantBase *)servant)->_private)->class_info = (gpointer)&class_info;
 }
 
 void
 POA_PortableServer_ServantActivator__fini(PortableServer_Servant servant,
-					  CORBA_Environment * ev)
+					CORBA_Environment * ev)
 {
-   PortableServer_ServantBase__fini(servant, ev);
+	PortableServer_ServantBase__fini(servant, ev);
 }
 
-void 
+void
 POA_PortableServer_ServantLocator__init(PortableServer_Servant servant,
 					CORBA_Environment * ev)
 {
-   static const PortableServer_ClassInfo class_info =
-   {NULL,
-    "IDL:omg.org/PortableServer/ServantLocator:1.0",
-    NULL };
+	static const PortableServer_ClassInfo class_info =
+	{NULL,
+	"IDL:omg.org/PortableServer/ServantLocator:1.0",
+	NULL};
 
-   PortableServer_ServantBase__init(((PortableServer_ServantBase *) servant), ev);
+	PortableServer_ServantBase__init(((PortableServer_ServantBase *)servant), ev);
 
-   ORBIT_OBJECT_KEY(((PortableServer_ServantBase *)servant)->_private)->class_info = (gpointer)&class_info;
+	ORBIT_OBJECT_KEY(((PortableServer_ServantBase *)servant)->_private)->class_info = (gpointer)&class_info;
 }
 
-void 
+void
 POA_PortableServer_ServantLocator__fini(PortableServer_Servant servant,
 					CORBA_Environment * ev)
 {
-   PortableServer_ServantBase__fini(servant, ev);
+	PortableServer_ServantBase__fini(servant, ev);
 }
+
+/* POA-related DSI stuff */
+static void
+dynamic_impl_skel(PortableServer_DynamicImpl *_ORBIT_servant,
+		  GIOPRecvBuffer *_ORBIT_recv_buffer,
+		  CORBA_Environment *ev,
+		  PortableServer_DynamicImplRoutine invoke)
+{
+	/* here the magic occurs... */
+	struct CORBA_ServerRequest_type sr;
+
+	ORBit_pseudo_object_init(ORBIT_PSEUDO_OBJECT(&sr),
+				 ORBIT_PSEUDO_SERVERREQUEST, ev);
+
+	CORBA_Object_duplicate((CORBA_Object)&sr, ev); /* just to make
+							  sure it doesn't die
+							  elsewhere */
+
+	sr.rbuf = _ORBIT_recv_buffer;
+	sr.orb = GIOP_MESSAGE_BUFFER(_ORBIT_recv_buffer)->connection->orb_data;
+
+	_ORBIT_servant->vepv->PortableServer_DynamicImpl_epv->invoke(_ORBIT_servant,
+								     &sr);
+
+	if(sr.sbuf) {
+		int i;
+		for(i = 0; i < sr.params->list->len; i++) {
+			CORBA_NamedValue *cur;
+			
+			cur = &g_array_index(sr.params->list, CORBA_NamedValue, i);
+			
+			if(cur->arg_modes & CORBA_ARG_IN) continue;
+
+			ORBit_marshal_arg(sr.sbuf, cur->argument._value,
+					  cur->argument._type);
+		}
+
+		giop_send_buffer_write(sr.sbuf);
+		giop_send_buffer_unuse(sr.sbuf);
+	} else
+		g_warning("Yo, your DSI code is messed up! You forgot to set_result|set_exception");
+
+	CORBA_NVList_free(sr.params, ev);
+}
+
+static ORBitSkeleton 
+dynamic_impl_get_skel(PortableServer_DynamicImpl * servant,
+		      GIOPRecvBuffer * _ORBIT_recv_buffer,
+		      gpointer * impl)
+{
+	*impl = (gpointer)servant->vepv->PortableServer_DynamicImpl_epv->invoke;
+
+	return (ORBitSkeleton)dynamic_impl_skel;
+}
+
+void
+PortableServer_DynamicImpl__init(PortableServer_Servant servant,
+				 CORBA_Environment *ev)
+{
+	static const PortableServer_ClassInfo class_info =
+	{(ORBitSkeleton (*)(PortableServer_ServantBase *, gpointer, gpointer *))
+	 &dynamic_impl_get_skel, "IDL:omg.org/CORBA/Object:1.0", NULL};
+	
+	PortableServer_ServantBase__init(servant, ev);
+	
+	ORBIT_OBJECT_KEY(((PortableServer_ServantBase *)servant)->_private)->class_info =
+		(PortableServer_ClassInfo *) & class_info;
+	
+}
+
+void PortableServer_DynamicImpl__fini(PortableServer_Servant servant,
+				      CORBA_Environment *ev)
+{
+	PortableServer_ServantBase__fini(servant, ev);
+}
+
