@@ -99,8 +99,9 @@ cbe_tc_generate(CBETCGenInfo *tci)
     subtypes_id = random_id++, sublabels_id = random_id++;
 
   if(strncmp(tci->structname, "anon", 4)) {
-    fprintf(tci->of, "#if (TC_IMPL_%s == ORBIT_FILE_ID_%s) && !defined(TC_DEF_%s)\n",
-	    tci->structname, tci->ci->c_base_name, tci->structname);
+    fprintf(tci->of, "#if ");
+    orbit_cbe_id_cond_hack(tci->of, "TC_IMPL", tci->structname, tci->ci->c_base_name);
+    fprintf(tci->of, " && !defined(TC_DEF_%s)\n", tci->structname);
     fprintf(tci->of, "#define TC_DEF_%s 1\n", tci->structname);
   }
 
@@ -396,7 +397,7 @@ cbe_tc_generate(CBETCGenInfo *tci)
 	  n++;
 	} else {
 	  /* default case */
-	  fprintf(tci->of, "{(CORBA_TypeCode)&TC_CORBA_octet_struct, (int *)&zero_int, CORBA_FALSE}");
+	  fprintf(tci->of, "{(CORBA_TypeCode)&TC_CORBA_octet_struct, (int *)&ORBit_zero_int, CORBA_FALSE}");
 	}
 
 	if(IDL_LIST(curlabel).next || IDL_LIST(curitem).next)
@@ -656,20 +657,20 @@ cbe_tc_generate(CBETCGenInfo *tci)
   switch(IDL_NODE_TYPE(tci->ts)) {
   case IDLN_TYPE_ENUM:
     if(IDL_TYPE_ENUM(tci->ts).enumerator_list)
-      fprintf(tci->of, "(char **)anon_subnames_array%d", subnames_id);
+      fprintf(tci->of, "(const char **)anon_subnames_array%d", subnames_id);
     else
       fprintf(tci->of, "NULL");
     break;
   case IDLN_TYPE_UNION:
     if(IDL_TYPE_UNION(tci->ts).switch_body)
-      fprintf(tci->of, "(char **)anon_subnames_array%d", subnames_id);
+      fprintf(tci->of, "(const char **)anon_subnames_array%d", subnames_id);
     else
       fprintf(tci->of, "NULL");
     break;
   case IDLN_TYPE_STRUCT:
   case IDLN_EXCEPT_DCL:
     if(IDL_TYPE_STRUCT(tci->ts).member_list)
-      fprintf(tci->of, "(char **)anon_subnames_array%d", subnames_id);
+      fprintf(tci->of, "(const char **)anon_subnames_array%d", subnames_id);
     else
       fprintf(tci->of, "NULL");
     break;
